package de.tud.kom.p2psim.api.resourcedistribution;

import de.tud.kom.p2psim.api.common.SimHostComponent;
import de.tud.kom.p2psim.api.energy.EnergyComponent;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.DistributableResource.RESOURCE_TYPE;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.resource.ResourceSensor;

/**
 * A variant of an {@link de.tud.kom.p2psim.api.energy.EnergyModel} that manages energy levels separately for different
 * kinds of resource units and allows to refill energy.
 * <p/>
 * You can access the ResourceModel component using {@link de.tud.kom.p2psim.api.common.SimHost#getComponent(Class)}.
 *
 * @author Tim Feuerbach
 */
public interface ResourceModel extends SimHostComponent {

    /**
     * Returns information an oracle {@link ResourceSensor} would provide to the associated host, particularly
     * the unit count per resource type.
     */
    ResourceSensor getResourceInformation();

    /**
     * Adds a new resource type to the model
     *
     * @param resourceTypeId a unique identifier; it is recommended to start from 0
     * @param name           a name for the resource type, only used for logs and visualizations
     * @param maxValue       the maximum value for this resource or -1 if there is no limit
     * @param initialValue   with how many resources the node should start
     * @param vital          whether the resource is vital to the node's existence. Depending on the scenario, nodes
     *                       whose vital
     *                       resources have been reduced to zero should be removed
     */
    public void registerResourceType(RESOURCE_TYPE resourceType, String name, int maxValue, int initialValue, boolean vital);

    /**
     * Adds a new listener that gets notified about changes in the resource stock.
     *
     * @param listener
     */
    public void addResourceChangeListener(ResourceChangeListener listener);

    /**
     * Removes the given listener.
     *
     * @return whether this listener was registered in the first place
     */
    public boolean removeResourceChangeListener(ResourceChangeListener listener);

    /**
     * Add a new consumer component to this hosts resource model for the given resource type.
     * <p/>
     * Note to implementors: Remember to register an {@link de.tud.kom.p2psim.api.energy.EnergyEventListener} to
     * react to resource consumptions by the component.
     *
     * @throws IllegalArgumentException if the given resource type has not been previously registered
     */
    public void registerComponent(EnergyComponent comp, RESOURCE_TYPE resourceType);

    /**
     * Increases the number of stored resource units for the given resource type up to the maximum capacity
     *
     * @param resourceTypeId identifier of the resource you want to fill up
     * @param resourceUnits  a positive value
     * @throws IllegalArgumentException if the resource type ID has not been registered beforehand
     * @throws IllegalArgumentException if resourceUnits is below 0
     */
    public void fillResourceUnits(RESOURCE_TYPE resourceType, double resourceUnits);
}
