package de.tud.kom.p2psim.impl.loadgen.config;

import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.HashSet;

import de.tud.kom.p2psim.impl.loadgen.exceptions.SiSTypeNotFoundException;
import de.tud.kom.p2psim.impl.loadgen.exceptions.UnknownTypeException;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.graph.INodeID;
import de.tudarmstadt.maki.simonstrator.api.component.sis.type.SiSTypes;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;

/**
 * A loader for the RequestType
 * 
 * @author Marc Schiller
 * @version 1.0, May 30, 2016
 */
public class RequestType {

	// Storage
	private String attribute;

	private String type;

	private long interval;

	private long validity;

	private HashSet<Target> targets = new HashSet<>();

	// TODO: Set the source of this request
	private INodeID source;

	// The list of Request Types
	public static final String[] REQUEST_TYPES = { "PERIODIC", "ONE-SHOT",
			"EVENT" };

	/**
	 * Constructor for (periodic) requests. If it should be one-shot, pass
	 * nothing for "interval" and "validity".
	 * 
	 * @param attribute
	 *            the attribute that should be requested
	 * @param type
	 *            the type of the request (== ONE-SHOT)
	 * @throws SiSTypeNotFoundException
	 *             if the wanted SiSType does not exist
	 * @throws UnknownTypeException
	 *             if the comparator does not exist
	 */
	@XMLConfigurableConstructor({ "attribute", "type" })
	public RequestType(String attribute, String type)
			throws SiSTypeNotFoundException, UnknownTypeException {
		this.setAttribute(attribute);
		this.setType(type);
	}

	/**
	 * Set the Attribute that should be requested
	 * 
	 * @param attribute
	 *            The SiS type
	 * @throws SiSTypeNotFoundException
	 *             if the wanted SiSType does not exist
	 */
	private void setAttribute(String attribute)
			throws SiSTypeNotFoundException {
		for (Field field : SiSTypes.class.getFields()) {
			if (field.getName().equals(attribute)) {
				this.attribute = attribute;
				break;
			}
		}
		if (this.attribute == null) {
			throw new SiSTypeNotFoundException();
		}
	}

	/**
	 * Set the type of the request (One-Shot, Periodic, Event)
	 * 
	 * @param type
	 * @throws UnknownTypeException
	 */
	private void setType(String type) throws UnknownTypeException {
		if (Arrays.asList(REQUEST_TYPES).contains(type)) {
			this.type = type;
		} else {
			throw new UnknownTypeException();
		}
	}

	/**
	 * Set the interval for periodic requests
	 * 
	 * @param interval
	 *            as time defined in {@link Time}
	 */
	public void setInterval(String interval) {
		this.interval = Time.parseTime(interval);
	}

	/**
	 * Set the validity for periodic requests
	 * 
	 * @param validtiy
	 *            as time defined in {@link Time}
	 */
	public void setValidity(String validity) {
		this.validity = Time.parseTime(validity);
	}

	/**
	 * Add a target to the list of targets (the nodes from whom data is
	 * requested)
	 * 
	 * @param target
	 */
	public void setTarget(Target target) {
		this.targets.add(target);
	}

	public String getAttribute() {
		return attribute;
	}

	public String getType() {
		return type;
	}

	public long getInterval() {
		return interval;
	}

	public long getValidity() {
		return validity;
	}

	public HashSet<Target> getTargets() {
		return targets;
	}

	public INodeID getSource() {
		return source;
	}

	/**
	 * Pretty print
	 */
	public String toString() {
		String retString = "Attribute: " + this.attribute + " / Type: "
				+ this.type;
		switch (this.type) {
		case "PERIODIC":
			retString += " / Validity: " + this.validity + " / Interval: "
					+ this.interval;
			break;
		case "ONE-SHOT":
			retString += retString;
			break;
		case "EVENT":
			retString += "";
			break;
		default:
			// Should never happen
			return "Error";
		}

		retString += "\n";

		for (Target tgt : this.targets) {
			retString += "\t\tTarget: " + tgt.toString() + "\n";
		}

		return retString;
	}

	// TODO: Handle Eventbased config
}
