package de.tud.kom.p2psim.impl.resourcedistribution;

import de.tud.kom.p2psim.api.analyzer.ResourceAnalyzer;
import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.api.energy.EnergyComponent;
import de.tud.kom.p2psim.api.energy.EnergyEventListener;
import de.tud.kom.p2psim.api.energy.EnergyState;
import de.tud.kom.p2psim.api.network.SimNetInterface;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceChangeListener;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceModel;
import de.tudarmstadt.maki.simonstrator.api.Monitor;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.core.MonitorComponent;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumer;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.ResourceConsumptionState;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.DistributableResource;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.DistributableResource.RESOURCE_TYPE;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.resource.ResourceSensor;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class DefaultResourceModel implements ResourceModel, ResourceSensor, EnergyEventListener {
    private SimHost host;
    private Map<RESOURCE_TYPE, DistributableResource> resourceInformation;
    private Set<ResourceChangeListener> resourceChangeListeners;

    private Map<EnergyComponent, RESOURCE_TYPE> energyComponents = new HashMap<>();
    private ResourceConsumer consumer;

    public DefaultResourceModel(SimHost host) {
        this.host = host;
        resourceInformation = new HashMap<RESOURCE_TYPE, DistributableResource>();
        resourceChangeListeners = new HashSet<>();
    }

    @Override
    public void initialize() {
        try {
            this.consumer = host.getComponent(ResourceConsumer.class);
        } catch (ComponentNotAvailableException e) {
            throw new RuntimeException("Resource model requires a resource consumer component on host", e);
        }

        try {
            ResourceAnalyzer resourceAnalyzer = Monitor.get(ResourceAnalyzer.class);
            resourceAnalyzer.initialize(consumer, this);
        } catch (MonitorComponent.AnalyzerNotAvailableException e) {
            // no analyzer present
        }
    }

    @Override
    public ResourceSensor getResourceInformation() {
        return this;
    }
 

    @Override
    public void registerResourceType(RESOURCE_TYPE resourceType, String name, int maxValue, int initialValue, boolean vital) {
        if (resourceInformation.containsKey(resourceType)) return;

        DistributableResource resource = new DefaultResourceContainer(resourceType, name, maxValue, initialValue, vital);         
        resourceInformation.put(resourceType, resource);
    }

    @Override
    public void addResourceChangeListener(ResourceChangeListener listener) {
        resourceChangeListeners.add(listener);
    }

    @Override
    public boolean removeResourceChangeListener(ResourceChangeListener listener) {
        return resourceChangeListeners.remove(listener);
    }

    @Override
    public void registerComponent(EnergyComponent comp, RESOURCE_TYPE resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            throw new IllegalArgumentException("Unknown resource " + resourceType.toString());
        }
        comp.setEnergyEventListener(this);
        energyComponents.put(comp, resourceType);
    }

    @Override
    public void fillResourceUnits(RESOURCE_TYPE resourceType, double resourceUnits) {
        if (!resourceInformation.containsKey(resourceType)) {
        		throw new IllegalArgumentException("Unknown resource " + resourceType.toString());
        }
        if (resourceUnits < 0) {
            throw new IllegalArgumentException(String.format("Number of resource units must be positive, %f given",
                    resourceUnits));
        }

        resourceInformation.get(resourceType).fill(resourceUnits);
    }

    @Override
    public void _fillResourceUnits(RESOURCE_TYPE resourceType, double resourceUnits) {
        fillResourceUnits(resourceType, resourceUnits);
    }
    
    @Override
	public void setResourcePercentage(RESOURCE_TYPE resourceType, double resourcePercentage) {
	    	if (!resourceInformation.containsKey(resourceType)) {
	    		throw new IllegalArgumentException("Unknown resource " + resourceType.toString());
	    }
	    	
	    	if (resourcePercentage < 0 || resourcePercentage > 1 ) {
	    		throw new IllegalArgumentException(String.format("Number of resource units must be between 0 and 1, %f given", resourcePercentage));
		}

	    	resourceInformation.get(resourceType).setPercentagedAmount(resourcePercentage);
	}

    @Override
    public void shutdown() {
        // not applicable
        // TODO find out expected behaviour
    }

    @Override
    public SimHost getHost() {
        return host;
    }

    @Override
    public List<RESOURCE_TYPE> getResourceTypes() {
    		List<RESOURCE_TYPE> resourceTypes = new LinkedList<RESOURCE_TYPE>();
		for(RESOURCE_TYPE res : resourceInformation.keySet())
		{
			resourceTypes.add(res);
		}
		
		return resourceTypes;
    }

    @Override
    public String getName(RESOURCE_TYPE resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            return null;
        }
        return resourceInformation.get(resourceType).toString();
    }

    @Override
    public double getResourceCount(RESOURCE_TYPE resourceType) {
        checkExistence(resourceType);
        return resourceInformation.get(resourceType).getAmount();
    }

    @Override
    public double getMaxResourceCount(RESOURCE_TYPE resourceType) {
        checkExistence(resourceType);
        return resourceInformation.get(resourceType).getMaxAmount();
    }

    private void checkExistence(RESOURCE_TYPE resourceType) {
        if (!resourceInformation.containsKey(resourceType)) {
            throw new IllegalArgumentException("Unknown resource type " + resourceType);
        }
    }

    @Override
    public void switchedState(EnergyComponent component, EnergyState oldState, EnergyState newState, long
            timeSpentInOldState) {
        DistributableResource resource = resourceInformation.get(energyComponents.get(component));     
        
        resource.extract((int) (oldState.getEnergyConsumption() *
                ((double) timeSpentInOldState / (double) Time.SECOND)));
             
        if (resource.isEmpty()) {
            // node is empty, should go offline
            for (SimNetInterface net : getHost().getNetworkComponent().getSimNetworkInterfaces()) {
                net.goOffline();
            }
            consumer.setState(ResourceConsumptionState.OFFLINE);
        }
    }

    @Override
    public boolean turnOn(EnergyComponent component) {
        return !resourceInformation.get(energyComponents.get(component)).isEmpty();
    }

    private void notifyListeners(DistributableResource resource, double oldValue) {
        for (ResourceChangeListener listener : resourceChangeListeners) {
            listener.resourceUnitsChanged(getHost(), resource.getResourceType(), oldValue, resource.getAmount());
        }
    }

    

	
}
