package de.tud.kom.p2psim.impl.topology.movement.resourceaware;

import de.tud.kom.p2psim.api.common.SimHost;
import de.tud.kom.p2psim.api.common.SimHostComponent;
import de.tud.kom.p2psim.api.resourcedistribution.ResourceModel;
import de.tud.kom.p2psim.api.topology.TopologyComponent;
import de.tud.kom.p2psim.api.topology.movement.SimLocationActuator;
import de.tudarmstadt.maki.simonstrator.api.Host;
import de.tudarmstadt.maki.simonstrator.api.Monitor;
import de.tudarmstadt.maki.simonstrator.api.Oracle;
import de.tudarmstadt.maki.simonstrator.api.Time;
import de.tudarmstadt.maki.simonstrator.api.common.UniqueID;
import de.tudarmstadt.maki.simonstrator.api.component.ComponentNotAvailableException;
import de.tudarmstadt.maki.simonstrator.api.component.core.MonitorComponent;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.*;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.DistributableResource.RESOURCE_TYPE;
import de.tudarmstadt.maki.simonstrator.api.component.resourcedistribution.analyzer.ResourceReceptionAnalyzer;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.Location;
import de.tudarmstadt.maki.simonstrator.api.component.sensor.location.LocationListener;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Monitors the movement of a single {@link ResourceConsumer} and relocates resources if it gets close to a targeted
 * {@link ResourceDistributor}.
 * <p/>
 * NOTE: Advertisements are not automatically updated. If you want to do this, listen for
 * {@link ResourceCollectionListener listener} events and update the store yourself.
 * <p/>
 * How close a node needs to be to collect a resource can be set with {@link #setDistanceToCollect(double)}; it defaults
 * to 1 meter.
 *
 * @author Tim Feuerbach
 * @see ResourceCollectionMovementMonitorFactory
 */
public class ResourceCollectionMovementMonitor implements SimHostComponent, LocationListener {
    private final SimHost host;
    double distanceToCollect = 1.0;

    protected ResourceConsumer consumer;
    protected ResourceModel resourceModel;
    protected TopologyComponent topology;
    protected List<ResourceCollectionListener> listeners;
    protected ResourceReceptionAnalyzer analyzer;

    protected final Map<UniqueID, ResourceDistributor> distributors = new HashMap<>();
    protected boolean distributorsInitialized = false;

    public ResourceCollectionMovementMonitor(SimHost host) {
        this.host = host;
    }

    @Override
    public void initialize() {
        if (!Oracle.isSimulation()) {
            throw new IllegalStateException("Resource collection can only be simulated when ran inside a simulation!");
        }

        try {
            consumer = host.getComponent(ResourceConsumer.class);
            resourceModel = host.getComponent(ResourceModel.class);
            topology = host.getTopologyComponent();
        } catch (ComponentNotAvailableException e) {
            throw new IllegalArgumentException(getClass().getName() + " missing required component!", e);
        }

        try {
            listeners = host.getComponents(ResourceCollectionListener.class);
            analyzer = Monitor.get(ResourceReceptionAnalyzer.class);
        } catch (ComponentNotAvailableException | MonitorComponent.AnalyzerNotAvailableException e) {
            // ok, no listener/analyzer configured
        }

        topology.requestLocationUpdates(null, this); // N.B.: using "null" leads to update on every location change
    }

    protected void initializeDistributorList() {
        for (Host host : Oracle.getAllHosts()) {
            try {
                ResourceDistributor distributor = host.getComponent(ResourceDistributor.class);
                distributors.put(distributor.getBeaconID(), distributor);
                distributor.registerListener(new ResourceDistributorListener() {                	
                		@Override
                    public void availableResourceAmountChanged(RESOURCE_TYPE resourceType, double oldValue, double newValue) {
                        // pass
                    }

                    @Override
                    public void beaconIdChanged(UniqueID oldId, UniqueID newId) {
                        synchronized (distributors) {
                            distributors.remove(oldId);
                            distributors.put(newId, distributor);
                        }
                    }
                });
            } catch (ComponentNotAvailableException e) {
                // ok, not a distributor
            }
        }

        if (distributors.isEmpty()) {
            Monitor.log(getClass(), Monitor.Level.WARN, "No implementors of " + ResourceDistributor.class
                    .getCanonicalName() + " are available in this simulation; resource collection will not work");
        }

        distributorsInitialized = true;
    }

    @Override
    public void onLocationChanged(Host host, Location location) {
        if (!consumer.getState().equals(ResourceConsumptionState.HEADING_FOR_RESOURCE)) return;

        if (!distributorsInitialized) {
            initializeDistributorList();
        }

        if (location.distanceTo(consumer.getTarget()) <= distanceToCollect) {
            // ensure we are actually at the correct distributor (can't trust the node!)
            synchronized (distributors) {
                ResourceDistributor distributor = distributors.get(consumer.getTargetBeaconID());
                if (distributor != null) {
                    try {
                        if (distributor.getHost().getComponent(SimLocationActuator.class).getRealPosition().distanceTo
                                (location) <= distanceToCollect) {
                        	
	                        	for(DistributableResource resource : distributor.getResources().getListOfResources())
	                        	{
	                        		 transferResourcesFromDistributor(distributor, resource.getResourceType());
	                        	}                       	
                           
                            return;
                        }
                    } catch (ComponentNotAvailableException e) {
                        throw new IllegalStateException("Cannot retrieve distributor's actual location: does not " +
                                "expose " +
                                "SimLocationActuator interface. Did you change DefaultTopologyComponent?", e);
                    }
                }
                //Monitor.log(ResourceCollectionMovementMonitor.class, Monitor.Level.WARN, "Beacon " +
                //        consumer.getTargetBeaconID() + " was not present at the position the node went to");
                beaconNotAvailable(consumer.getTargetBeaconID());
            }
        }
    }

    protected void transferResourcesFromDistributor(ResourceDistributor distributor, RESOURCE_TYPE resourceType) {
        double resourcesGot = distributor.takeResources(resourceType, consumer.getDemand());
        resourceModel.fillResourceUnits(resourceType, resourcesGot);
        if (analyzer != null) {
            analyzer.visitedResourceDistributor(consumer.getHost().getId(), distributor.getBeaconID(), resourceType, resourcesGot, Time.getCurrentTime());
        }
        notifyListener(distributor.getBeaconID(), distributor.getBeaconSequenceID(), resourcesGot, distributor.getResourceUnits(resourceType));
    }

    protected void beaconNotAvailable(UniqueID beaconId) {
        notifyListener(beaconId, Integer.MAX_VALUE, 0, ResourceCollectionListener.BEACON_NOT_AVAILABLE);
    }

    /**
     * Notifies the host's resource collection listener (if existing). Method can be overridden to conceal the actual
     * amount of resources available at the distributor (so the node has to wait for a new announcement).
     */
    protected void notifyListener(UniqueID beaconID, int beaconSequenceId, double resourcesGot, double resourcesRemaining) {
        if (listeners != null) {
            for (ResourceCollectionListener listener : listeners) {
                listener.resourcesCollected(beaconID, beaconSequenceId, resourcesGot, resourcesRemaining);
            }
        }
    }

    @Override
    public void shutdown() {
        // no action necessary
    }

    @Override
    public SimHost getHost() {
        return host;
    }

    /**
     * Defines how far away a node can be to collect a resource.
     */
    public void setDistanceToCollect(double distanceToCollect) {
        this.distanceToCollect = distanceToCollect;
    }
}
